print("import: 'xrayutilities'")
import xrayutilities

print("import: 'xrayutilities.analysis'")
import xrayutilities.analysis

print("import: 'xrayutilities.io'")
import xrayutilities.io

print("import: 'xrayutilities.materials'")
import xrayutilities.materials

print("import: 'xrayutilities.math'")
import xrayutilities.math

print("import: 'xrayutilities.simpack'")
import xrayutilities.simpack

